/*
 * Copyright (c) 2010, Stanislav Muhametsin. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package org.atp.spi;

import java.util.HashMap;
import java.util.Map;

import org.atp.api.Typeable;

/**
 * <p>
 * This is skeleton implementation for objects, which store mapping between each API type and interaction. Typically,
 * when context consists of only one object, the interaction is based at least on type of object. This class is meant to
 * provide clear and reliable implementation for storing interaction for each API type.
 * </p>
 * 
 * <p>
 * Please note that this class is not thread-safe. Additionally, since there is only one interaction object for each
 * <b>type</b> of object, interactions are good to be stateless objects.
 * </p>
 * 
 * @author Stanislav Muhametsin
 * @param <BaseType> The base type for all objects, for which there may exist interaction.
 * @param <InteractionType> The base type for all interactions.
 */
public class InteractionMapper<BaseType extends Typeable<?>, InteractionType>
{

    private final Map<Class<? extends BaseType>, InteractionType> _mapping;

    /**
     * The strategy type for situations when user tries to associate interaction for specific API type.
     * 
     * @author Stanislav Muhametsin
     * @param <InteractionType> The type of interaction.
     */
    public static interface AlreadyContainsInteractionStrategy<InteractionType>
    {
        /**
         * This method will be called when user tries to associate interaction for specific API type.
         * 
         * @param type The API type of object.
         * @param interaction The interaction to be associated with the type.
         * @param containedInteraction The interaction already associated with the type.
         */
        public void alreadyContainsInteraction( Class<? extends Typeable<?>> type, InteractionType interaction,
            InteractionType containedInteraction );
    }

    /**
     * The strategy type for situations, when user ties to retrieve interaction for object and there is no interaction
     * found for the API type of the object.
     * 
     * @author Stanislav Muhametsin
     * @param <InteractionType> The type of interaction.
     */
    public static interface DoesNotContainInteractionStrategy<InteractionType>
    {
        /**
         * This method will be called when user tries to retrieve interaction and there is no interaction found for the
         * API type of the object.
         * 
         * @param object The object for which user tries to retrieve interaction.
         * @return If possible, should return some "default" or "no-op" interaction. Otherwise, null should be returned
         *         or exception should be thrown.
         */
        public InteractionType doesNotContainInteraction( Typeable<?> object );
    }

    /**
     * The strategy, which will throw {@link IllegalArgumentException} when its method is called.
     */
    public static final AlreadyContainsInteractionStrategy<Object> THROW_IF_CONTAINS = new AlreadyContainsInteractionStrategy<Object>()
    {

        /**
         * Immediately throws {@link IllegalArgumentException}.
         * 
         * @throws IllegalArgumentException Always.
         */
        public void alreadyContainsInteraction( Class<? extends Typeable<?>> type, Object interaction,
            Object containedInteraction )
        {
            throw new IllegalArgumentException( "Interaction mapper already contained interaction for "
                + type.getName() + ". The contained interaction is "
                + (containedInteraction == null ? "[null]" : containedInteraction.toString()) + "." );
        }
    };

    /**
     * The strategy, which will throw {@link IllegalArgumentException} when its method is called.
     */
    public static final DoesNotContainInteractionStrategy<Typeable<?>> THROW_IF_NOT_CONTAINS = new DoesNotContainInteractionStrategy<Typeable<?>>()
    {

        /**
         * Immediately throws {@link IllegalArgumentException}.
         * 
         * @throws IllegalArgumentException Always.
         */
        public Typeable<?> doesNotContainInteraction( Typeable<?> object )
        {
            throw new IllegalArgumentException( "Interaction mapper does not contain interaction for "
                + (object == null ? "[null]" : object.getImplementedType()) + "." );
        }
    };

    // @formatter:off
    /**  
     * Constructs empty instance of interaction classifier. Calling this constructor is equivalent on calling
     * {@link InteractionMapper(new HashMap())}
     * 
     * @see InteractionMapper#InteractionMapper(Map)
     */
    // @formatter:on
    protected InteractionMapper()
    {
        this( new HashMap<Class<? extends BaseType>, InteractionType>() );
    }

    /**
     * Constructs an instance of interaction classifier, which uses the given map as initial mapping.
     * 
     * @param mapToUse The mapping to use. If it is {@code null}, the mapping will be initialized as empty instance of
     *            {@link HashMap}.
     */
    protected InteractionMapper( Map<Class<? extends BaseType>, InteractionType> mapToUse )
    {
        if( mapToUse == null )
        {
            this._mapping = new HashMap<Class<? extends BaseType>, InteractionType>();
        }
        else
        {
            this._mapping = mapToUse;
        }
    }

    /**
     * Tries to associate some interaction for some API type of object. Calling this method is equivalent to calling
     * {@link #associate(type, interaction, THROW_IF_CONTAINS)}
     * 
     * @param type The API type of object.
     * @param interaction The interaction for this type of objects.
     * @throws IllegalArgumentException If another interaction is already assigned for {@code type}.
     * @see #THROW_IF_CONTAINS
     */
    @SuppressWarnings("unchecked")
    protected void associate( Class<? extends BaseType> type, InteractionType interaction )
    {
        // Can cast safely, since any parameter may be casted to Object.
        this.associate( type, interaction, (AlreadyContainsInteractionStrategy<InteractionType>) THROW_IF_CONTAINS );
    }

    /**
     * Tries to associate some interaction for some API type of object. Will use given strategy (if any) if a mapping
     * already contains interaction (it checks it through {@link Map#containsKey(Object)} method).
     * 
     * @param type The API type of object. May not be null (check is not performed).
     * @param interaction The interaction for this type of objects. May be null (check is not performed).
     * @param strategy The strategy when there already exists interaction for given type. May be null (check is
     *            performed).
     */
    protected void associate( Class<? extends BaseType> type, InteractionType interaction,
        AlreadyContainsInteractionStrategy<InteractionType> strategy )
    {
        if( this._mapping.containsKey( type ) )
        {
            if( strategy != null )
            {
                strategy.alreadyContainsInteraction( type, interaction, this._mapping.get( type ) );
            }
        }
        else
        {
            this._mapping.put( type, interaction );
        }
    }

    /**
     * Tries to retrieve interaction associated with the API type of the given object. Calling this method is equivalent
     * to calling {@link #getInteraction(object, THROW_IF_NOT_CONTAINS)}.
     * 
     * @param object The object for which the interaction is wanted-
     * @return The interaction associated for the API type of object.
     * @throws IllegalArgumentException If no interaction is registered for the API type of object.
     * @see #THROW_IF_NOT_CONTAINS
     */
    @SuppressWarnings("unchecked")
    protected InteractionType getInteraction( BaseType object )
    {
        // Can cast safely as the strategy will never return value.
        return this.getInteraction( object, (DoesNotContainInteractionStrategy<InteractionType>) THROW_IF_NOT_CONTAINS );
    }

    /**
     * Tries to retrieve interaction associated with the API type of the given object. If there is no interaction
     * associated with this type, it will use given strategy (if any) to get interaction.
     * 
     * @param object The object for which retrieve interaction. May not be null.
     * @param strategy The strategy to use when no interaction is found (the {@link Map#get(Object)} method is used to
     *            check if there is interaction, giving the {@link Typeable#getImplementedType()} as argument). May be
     *            null (check is performed).
     * @return The interaction for this object. Return {@code null}, if there was no mapping or mapping to null, and
     *         given strategy returned null.
     * @throws NullPointerException If {@code object} is null.
     */
    protected InteractionType getInteraction( BaseType object,
        DoesNotContainInteractionStrategy<InteractionType> strategy )
    {
        InteractionType interaction = this._mapping.get( object.getImplementedType() );
        if( interaction == null && strategy != null )
        {
            interaction = strategy.doesNotContainInteraction( object );
        }

        return interaction;
    }

}
