/*
 * Copyright (c) 2010, Stanislav Muhametsin. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package org.atp.spi;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.atp.api.Typeable;

/**
 * Since {@link InteractionMapper} was not thread-safe, this class provides thread-safe implementations for mapping
 * between each API type and interaction.
 * 
 * @author Stanislav Muhametsin
 * @see InteractionMapper
 */
public class ThreadSafeInteractionMapper<BaseType extends Typeable<?>, InteractionType> extends
    InteractionMapper<BaseType, InteractionType>
{

    /**
     * The lock to use when accessing mapping.
     */
    private final ReadWriteLock _lock;

    /**
     * Constructs new instance with empty mapping and default lock. Equivalent on calling {@link
     * #ThreadSafeInteractionMapper(new HashMap())}
     * 
     * @see #ThreadSafeInteractionMapper(Map)
     */
    protected ThreadSafeInteractionMapper()
    {
        this( new HashMap<Class<? extends BaseType>, InteractionType>() );
    }

    /**
     * Constructs new instance with given mapping and default lock. Equivalent on calling {@link
     * #ThreadSafeInteractionMapper(mapToUse, new ReentrantReadWriteLock( true ))}.
     * 
     * @param mapToUse The mapping to use
     * @see #ThreadSafeInteractionMapper(Map, ReadWriteLock)
     */
    protected ThreadSafeInteractionMapper( Map<Class<? extends BaseType>, InteractionType> mapToUse )
    {
        // Construct fair lock by default, so that reading will be a little faster.
        this( mapToUse, new ReentrantReadWriteLock( true ) );
    }

    /**
     * Constructs new instance with empty mapping and given lock. Equivalent on calling {@link
     * #ThreadSafeInteractionMapper(new HashMap(), lockToUse)}.
     * 
     * @param lockToUse The lock to use.
     * @see #ThreadSafeInteractionMapper(Map, ReadWriteLock)
     */
    protected ThreadSafeInteractionMapper( ReadWriteLock lockToUse )
    {
        this( new HashMap<Class<? extends BaseType>, InteractionType>(), lockToUse );
    }

    /**
     * Constructs new instance with given mapping and given lock. If the lock is null, then it defaults to
     * {@link ReentrantReadWriteLock} with passing {@code true} to constructor
     * {@link ReentrantReadWriteLock#ReentrantReadWriteLock(boolean)}.
     * 
     * @param mapToUse The mapping to use. May be null.
     * @param lockToUse The lock to use. May be null.
     * @see InteractionMapper#InteractionMapper(Map)
     */
    protected ThreadSafeInteractionMapper( Map<Class<? extends BaseType>, InteractionType> mapToUse,
        ReadWriteLock lockToUse )
    {
        super( mapToUse );

        if( lockToUse == null )
        {
            this._lock = new ReentrantReadWriteLock( true );
        }
        else
        {
            this._lock = lockToUse;
        }
    }

    /**
     * Thread-safe implemenation for association interaction.
     */
    @Override
    protected void associate( Class<? extends BaseType> type, InteractionType interaction,
        AlreadyContainsInteractionStrategy<InteractionType> strategy )
    {
        this._lock.writeLock().lock();
        try
        {
            super.associate( type, interaction, strategy );
        }
        finally
        {
            this._lock.writeLock().unlock();
        }
    }

    /**
     * Thread-safe implementation for getting interaction.
     */
    @Override
    protected InteractionType getInteraction( BaseType object,
        DoesNotContainInteractionStrategy<InteractionType> strategy )
    {
        this._lock.readLock().lock();
        try
        {
            return super.getInteraction( object, strategy );
        }
        finally
        {
            this._lock.readLock().unlock();
        }
    }
}
